/*

  Copyright (C) 1999 President and Fellows of Harvard College

  All rights reserved.

  NOTICE: This software is provided "as is", without any warranty,
  including any implied warranty for merchantability or fitness for a
  particular purpose.  Under no circumstances shall Harvard University
  or its faculty, staff, students or agents be liable for any use of,
  misuse of, or inability to use this software, including incidental
  and consequential damages.

  License is hereby given to use, modify, and redistribute this
  software, in whole or in part, for any commercial or non-commercial
  purpose, provided that the user agrees to the terms of this
  copyright notice, including disclaimer of warranty, and provided
  that this copyright notice, including disclaimer of warranty, is
  preserved in the source code and documentation of anything derived
  from this software.  Any redistributor of this software or anything
  derived from this software assumes responsibility for ensuring that
  any parties to whom wuch a redistribution is made are fully aware of
  the terms of this license and disclaimer.

  Author: Brad Karp, Harvard University EECS, September, 1997

*/

/* arp.c -- ARP-related (IP-to-MAC) functionality for HUMR */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "debug.h"
#include "humr.h"
#include "arp.h"
#include "modem.h"

/* don't want to pull in all the loop.h stuff */
int humr_send_packet(u_char *, int, u_char *, int);

/* linked list head: all known handles */
static humr_handent *handent_head = NULL;

/* the hash table of NNARP entries */
static humr_nnarpent *nnarp_table[256];
/* active list of NNARP entries */
static humr_nnarpent *nnarpent_active = NULL;

/* my own MAC address */
humr_mac_addr mymac;
/* my own IP address */
struct in_addr myip;
/* XXX - fallback next hop IP */
struct in_addr fallback = { 0 };

/* time of last callout */
struct timeval lastcout;
/* approximate time of last event that hinted modem alive and in StarMode */
struct timeval keepalive;

/* hash an IP address down to one of 256 ARP table locations.
   convention: ALWAYS call with network-byte-order IP addresses!! */
unsigned char iphash(struct in_addr theip)
{
  unsigned char retval = 0;
  int i;

  for (i = 0; i < 4; i++, theip.s_addr >>= 8)
    retval ^= (theip.s_addr & 0xff);
  return retval;
}

/* initialize the NNARP table */
void nnarp_init(void)
{
  int i;

  for (i = 0; i < 256; i++)
    nnarp_table[i] = NULL;
}

/* add the given MAC-IP pair to the NNARP table */
int nnarp_add(humr_mac_addr mac, struct in_addr theip)
{
  humr_nnarpent *tmp;
  unsigned char loc;

  if ((tmp = (humr_nnarpent *) malloc(sizeof(humr_nnarpent))) != NULL) {
    tmp->ipdst = theip;
    tmp->macdst = mac;
    tmp->state = NNARPENT_REF0;
    tmp->actnext = nnarpent_active;
    nnarpent_active = tmp;
    loc = iphash(theip);
    tmp->next = nnarp_table[loc];
    nnarp_table[loc] = tmp;
    return (1);
  }
  return(0);
}

/* find ent in hash table, delete from its chain. return 1 if found and
   deleted; 0 otherwise. */
int nnarp_hashdelete(humr_nnarpent *ent)
{
  unsigned char loc;
  humr_nnarpent *hashhead, *hashprev;

  loc = iphash(ent->ipdst);
  hashhead = nnarp_table[loc];
  if (hashhead == NULL)
    return(0);
  if (hashhead == ent) {
    nnarp_table[loc] = hashhead->next;
    return(1);
  }
  else {
    hashprev = hashhead;
    hashhead = hashhead->next;
    while (hashhead != NULL) {
      if (hashhead == ent) {
	hashprev->next = hashhead->next;
	return(1);
      }
      hashprev = hashhead;
      hashhead = hashhead->next;
    }
  }
  return(0);
}

/* return pointer to handent if findhand found in handle list, otherwise
   return NULL. findhand MUST be NUL terminated. */
humr_handent *hand_lookup(char *findhand)
{
  humr_handent *here = handent_head;

  while (here != NULL) {
    if (strncmp(here->handle, findhand, here->handlen) == 0)
      /* found it */
      return(here);
    else
      here = here->next;
  }
  return(NULL);
}

/* yank handle from handle list if it's there; return 0 on failure, 1 on
   success */
int hand_delete(char *killhand)
{
  humr_handent *tmp = handent_head, *prev;

  while (tmp != NULL) {
    if (strncmp(tmp->handle, killhand, tmp->handlen) == 0) {
      /* found it */
      if (tmp == handent_head) {
	handent_head = handent_head->next;
	free(tmp);
      }
      else {
	prev->next = tmp->next;
	free(tmp);
      }
      return(1);
    }
    else {
      prev = tmp;
      tmp = tmp->next;
    }
  }
  return(0);
}

/* flush all entries from handle table */
void hand_flushall(void)
{
  humr_handent *tmp = handent_head;

  while (handent_head != NULL) {
    tmp = handent_head;
    handent_head = handent_head->next;
    free(handent_head);
  }
}

/* add newhand to the handle list */
int hand_add(char *newhand)
{
  humr_handent *tmp;

  if ((tmp = (humr_handent *) malloc(sizeof(humr_handent))) != NULL) {
    strncpy((char *)&tmp->handle, newhand, sizeof(tmp->handle));
    tmp->handlen = strlen(newhand);
    tmp->state = HANDENT_WAITING;
    tmp->next = handent_head;
    handent_head = tmp;
    return(1);
  }
  return(0);
}

/* called once per minute (roughly) */
void timertick(void)
{
  humr_handent *handent = handent_head;
  humr_handent *hprev;
  humr_handent *hkill = NULL;
  humr_nnarpent *nnarpent = nnarpent_active, *nnarpprev;
  humr_nnarpent *nnkill = NULL;

  /* for each handle in handle table */
  while (handent != NULL) {
    if (handent->state == HANDENT_WAITING) {
      /* re-send i-am, age entry */
      handent->state = HANDENT_OLD;
    }
    send_iam_handle(handent->handle, handent->handlen);
    hprev = handent;
#if 0
    }
    else if (handent->state == HANDENT_OLD) {
      /* delete old entry */
      if (handent == handent_head)
	handent_head = handent->next;
      else
	hprev->next = handent->next;
      hkill = handent;
    }
#endif
    handent = handent->next;
    if (hkill != NULL) {
      free(hkill);
      hkill = NULL;
    }
  }
#if 0
  /* now we only ask for neighbors at startup; use async INFO messages
     for discovery thereafter */
  /* ask modem for handle report */
  ask_neighbors();
#endif
  /* for each NNARP entry in NNARP table */
  while (nnarpent != NULL) {
    /* refresh remote table entry */
    send_iam_mac(nnarpent->macdst);
    switch(nnarpent->state) {
    case NNARPENT_REF0:
    case NNARPENT_REF1:
    case NNARPENT_REF2:
      /* advance refresh state */
      nnarpent->state++;
      nnarpprev = nnarpent;
      break;
    case NNARPENT_REF3:
      /* entry expired! delete */
      /* first, remove from active list */
      if (nnarpent == nnarpent_active)
	nnarpent_active = nnarpent->actnext;
      else
	nnarpprev->actnext = nnarpent->actnext;
      nnkill = nnarpent;
      /* now, remove from hash list */
      nnarp_hashdelete(nnarpent);
      break;
    default:
      break;
    }
    nnarpent = nnarpent->actnext;
    if (nnkill != NULL) {
      free(nnkill);
      nnkill = NULL;
    }
  }
}

/* add a new handle to the handle table, if it's not there already, and send
   an I-am to the handle. handle MUST be NUL-terminated. */
int add_neighbor(char *handle)
{
  if (hand_lookup(handle) == NULL) {
    send_iam_handle(handle, strlen(handle));
    hand_add(handle);
    return(1);
  }
  return(0);
}

/* heard a MAC-IP pair; put it in table as appropriate. src is the raw
   starmode source as a NUL-terminated string; we need it to see if source is
   still a handle.*/
int receive_iam(humr_mac_addr mac, struct in_addr theip, char *src)
{
  unsigned char loc;
  humr_nnarpent *here = NULL;

  if (dmask & DMASK_IAM) {
    fprintf(stderr, "heard i-am: IP %s is MAC %d-%d\n", inet_ntoa(theip),
	    ntohs(mac.hi), ntohs(mac.lo));
    fflush(stderr);
  }
  mac.hi = ntohs(mac.hi);
  mac.lo = ntohs(mac.lo);
  if (ishandle(src))
    hand_delete(src);
  loc = iphash(theip);
  if (nnarp_table[loc] != NULL) {
    here = nnarp_table[loc];
    while (here != NULL) {
      if (here->ipdst.s_addr == theip.s_addr) {
	if ((here->macdst.hi == mac.hi) && (here->macdst.lo == mac.lo))
	  here->state = NNARPENT_REF0;
	else {
	  fprintf(stderr, "replacing IP %ld's NNARP entry\n", theip.s_addr);
	  here->macdst = mac;
	  here->state = NNARPENT_REF0;
	}
	break;
      }
      here = here->next;
    }
  }
  if (here == NULL) {
    /* didn't find IP in ARP table */
    nnarp_add(mac, theip);
    send_iam_mac(mac);
  }
}

/* send an "I-am" to the peer with MAC address mac */
int send_iam_mac(humr_mac_addr mac)
{
  u_char modem_head[256];
  u_char payload[40];
  humr_header *outhead = (humr_header *) payload;

  modem_head[0] = '*';
  /* insert destination MAC addr */
  sprintf(&modem_head[1], "%04d", mac.hi);
  modem_head[5] = '-';
  sprintf(&modem_head[6], "%04d", mac.lo);
  modem_head[10] = '*';
  outhead->src.hi = htons(mymac.hi);
  outhead->src.lo = htons(mymac.lo);
  outhead->humr_type = htons(HUMR_TYPE_IAM);
  memcpy(payload+sizeof(humr_header), &myip, sizeof(struct in_addr));
  if (dmask & DMASK_IAM) {
    fprintf(stderr, "send i-am to mac %d-%d\n", mac.hi, mac.lo);
    fflush(stderr);
  }
  humr_send_packet(modem_head, 11,
		   payload, sizeof(humr_header)+sizeof(struct in_addr));
}

/* send an "I-am" to the peer with handle handle */
int send_iam_handle(char *handle, int handlen)
{
  u_char modem_head[256];
  u_char payload[40];
  humr_header *outhead = (humr_header *) payload;

  modem_head[0] = '*';
  /* insert destination handle addr */
  memcpy(&modem_head[1], handle, handlen);
  modem_head[1+handlen] = '*';
  outhead->src.hi = htons(mymac.hi);
  outhead->src.lo = htons(mymac.lo);
  outhead->humr_type = htons(HUMR_TYPE_IAM);
  memcpy(payload+sizeof(humr_header), &myip, sizeof(struct in_addr));
  if (dmask & DMASK_IAM) {
    char dhand[80];
    memcpy(dhand, handle, handlen);
    dhand[handlen] = '\0';
    fprintf(stderr, "send i-am to handle %s\n", dhand);
    fflush(stderr);
  }
  humr_send_packet(modem_head, 2+handlen,
		   payload, sizeof(humr_header)+sizeof(struct in_addr));
}

/* look up NNARP entry for a _network_-ordered IP address dip.
   return NULL if not found, pointer to NNARP entry if found. */
humr_nnarpent *nnarp_lookup(struct in_addr dip)
{
  humr_nnarpent *here;
  unsigned char loc;

  loc = iphash(dip);
  if (nnarp_table[loc] != NULL) {
    here = nnarp_table[loc];
    while (here != NULL) {
      if (here->ipdst.s_addr == dip.s_addr)
	return(here);
      here = here->next;
    }
  }
  /* XXX - if we can't find dest IP, fall back on default next hop IP */
  /* should this always be other end of tunnel? if so, fetch automatically. */
  if ((fallback.s_addr != 0) && (dip.s_addr != fallback.s_addr))
    return(nnarp_lookup(fallback));
  return(NULL);
}

static humr_nnarpent *curriter = NULL;

/* reset neighbor iteration */
void nnarp_neighb_reset(void)
{
  curriter = nnarpent_active;
}

/* iterate over all neighbors, returning one NNARP table entry for each.
   reset == non-zero resets iteration. */
humr_nnarpent *nnarp_neighb_iter(void)
{
  humr_nnarpent *retval;

  if (curriter == NULL)
    return(NULL);
  else {
    retval = curriter;
    curriter = curriter->actnext;
    return(retval);
  }
}
