/*

  Copyright (C) 1999 President and Fellows of Harvard College

  All rights reserved.

  NOTICE: This software is provided "as is", without any warranty,
  including any implied warranty for merchantability or fitness for a
  particular purpose.  Under no circumstances shall Harvard University
  or its faculty, staff, students or agents be liable for any use of,
  misuse of, or inability to use this software, including incidental
  and consequential damages.

  License is hereby given to use, modify, and redistribute this
  software, in whole or in part, for any commercial or non-commercial
  purpose, provided that the user agrees to the terms of this
  copyright notice, including disclaimer of warranty, and provided
  that this copyright notice, including disclaimer of warranty, is
  preserved in the source code and documentation of anything derived
  from this software.  Any redistributor of this software or anything
  derived from this software assumes responsibility for ensuring that
  any parties to whom wuch a redistribution is made are fully aware of
  the terms of this license and disclaimer.

  Author: Brad Karp, Harvard University EECS, September, 1997

*/

/* loop.c -- 
   main processing loop of HUMR, Harvard User-level Metricom Radio Driver */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <net/route.h>
#include "humr.h"
#include "modem.h"
#include "arp.h"
#include "stuff.h"

int tunnelfd;
int lflag = 0;
struct in_addr ignore = { INADDR_ANY };
int rtfd;

/* send a HUMR packet. pbuf MUST be large enough for the whole packet,
   including header, stuffed payload, and checksum. payidx is the index into
   pbuf where the stuffed payload should be written. payload contains
   the packet payload. paylen is the length of payload.
   this routine prepares the packet for sending in pbuf, allocated elsewhere
   (either by the caller or globally). */
int humr_send_packet(u_char *pbuf, int payidx, u_char *payload, int paylen)
{
  u_char *state = NULL;
  u_char *postpay;
  unsigned int cksum;

  /* stuff payload bytes into pbuf */
  postpay = StuffData(payload, paylen, pbuf+payidx, &state);
  /* compute checksum and append to packet */
  cksum = packet_checksum(pbuf, postpay-pbuf);
  if (cksum != CHECKSUM_INVALID)
    sprintf(postpay, "%04X", cksum);
  else
    fprintf(stderr, "can't checksum packet in humr_send_packet!\n");
  /* append trailing <cr> */
  postpay += 4;
  *postpay = '\r';
  /* write to modem ! */
  write(modemfd, pbuf, postpay-pbuf+1);
}

/* send packet to MAC addr macdst, with payload payload of length paylen */
int humr_ip_send(humr_nnarpent *macdst, u_char *payload, int paylen)
{
  u_char modem_head[2048];
  u_char outpay[1500];
  humr_header *outhead = (humr_header *) outpay;

  modem_head[0] = '*';
  /* insert destination MAC addr */
  sprintf(&modem_head[1], "%04d", macdst->macdst.hi);
  modem_head[5] = '-';
  sprintf(&modem_head[6], "%04d", macdst->macdst.lo);
  modem_head[10] = '*';
  outhead->src.hi = htons(mymac.hi);
  outhead->src.lo = htons(mymac.lo);
  outhead->humr_type = htons(HUMR_TYPE_IP);
  memcpy(outpay+sizeof(humr_header), payload, paylen);
  humr_send_packet(modem_head, 11, outpay, sizeof(humr_header)+paylen);
}

static struct rt_msghdr rtfd_head;
static struct iovec rtfd_iov[2];
static struct sockaddr_in rtfd_findme;

/* initialize kernel routing socket */
void rtfd_init(void)
{
  rtfd = socket(PF_ROUTE, SOCK_RAW, AF_INET);
  bzero(&rtfd_head, sizeof(rtfd_head));
  bzero(&rtfd_findme, sizeof(rtfd_findme));
  rtfd_findme.sin_family = AF_INET;
  rtfd_findme.sin_len = sizeof(struct sockaddr_in);
  rtfd_head.rtm_version = RTM_VERSION;
  /* we're looking up a route */
  rtfd_head.rtm_type = RTM_GET;
  /* which flags to set on new entry */
  rtfd_head.rtm_flags = RTF_UP | RTF_GATEWAY | RTF_HOST;
  /* which addresses we're enclosing in the message (LSB to MSB!) */
  rtfd_head.rtm_addrs = RTA_DST;
  /* originator (me) of the request */
  rtfd_head.rtm_pid = getpid();
  /* sequence number of the message */
  rtfd_head.rtm_seq = 0;
  /* set message length */
  rtfd_head.rtm_msglen = sizeof(struct rt_msghdr) + sizeof(struct sockaddr_in);
  /* set up gathered write */
  rtfd_iov[0].iov_base = (char *) &rtfd_head;
  rtfd_iov[0].iov_len = sizeof(rtfd_head);
}

struct in_addr rtfd_lookup(struct in_addr dst) {
  unsigned char inrtmsg[sizeof(struct rt_msghdr)+
		       sizeof(struct sockaddr)*RTAX_MAX];
  struct rt_msghdr *rthdr;
  struct sockaddr_in *ipaddr;
  int inlen;
  struct in_addr bogus;

  rtfd_findme.sin_addr = dst;
  rtfd_iov[1].iov_base = (char *) &rtfd_findme;
  rtfd_iov[1].iov_len = sizeof(struct sockaddr_in);
  writev(rtfd, rtfd_iov, 2);
  while (1) {
    inlen = read(rtfd, inrtmsg, sizeof(inrtmsg));
    rthdr = (struct rt_msghdr *) inrtmsg;
    if ((rthdr->rtm_pid == rtfd_head.rtm_pid) &&
	(rthdr->rtm_seq == rtfd_head.rtm_seq)) {
      rtfd_head.rtm_seq++;
      if ((rthdr->rtm_errno == 0) && (rthdr->rtm_flags & RTF_HOST) &&
	  !(rthdr->rtm_flags & RTF_WASCLONED)) {
	/* only see a match if the route is a host route, and there was
	   no error, and the route wasn't cloned */
	/* stuff gateway IP address into hdr->ip_dst */
	/* XXX we know we'll get a destination addr and a gateway addr, in that
	   order */
	ipaddr = (struct sockaddr_in *) (inrtmsg+sizeof(struct rt_msghdr)+
					 sizeof(struct sockaddr));
	return(ipaddr->sin_addr);
      }
      else {
	if (ignore.s_addr != INADDR_ANY) {
	  if (dst.s_addr != ignore.s_addr)
	    return(dst);
	  else {
	    struct in_addr dummy = { INADDR_ANY };
	    return(dummy);
	  }
	}
	else
	  return(dst);
      }
    }
  }    
}

/* endlessly pass packets from tunnel to serial line and vice versa */
void mainloop(void)
{
  fd_set rfds;
  static char paccum[65536];
  unsigned char recvpacket[2048];
  humr_header *recvheader = (humr_header *) recvpacket;
  unsigned char txpacket[2048];
  static char *prevptr = paccum;
  static char *accptr = paccum;
  char *curr;
  int ptype;
  int plen;
  int nbytes = 0;                     /* number of bytes read/written */
  static int buffered = 0;            /* number of chars buffered from modem */
  char srcaddr[20];                   /* source address of packet from modem */
  unsigned int csum;                  /* cksum computed */
  unsigned int rsum;                  /* cksum received */
  struct ip *ipee;                    /* IP header */
  struct timeval onesec;              /* select timeout */
  struct timeval now;                 /* current time */
  humr_nnarpent *macdest;             /* nnarp result for packet tx */

  onesec.tv_sec = 1;
  onesec.tv_usec = 0;
  while (1) {
    FD_ZERO(&rfds);
    FD_SET(tunnelfd, &rfds);
    FD_SET(modemfd, &rfds);
    /* block until a packet is presented at serial or tunnel */
    /* XXX -- what's max descriptor number? */
    select(8, &rfds, NULL, NULL, &onesec);
    /* check if it's time to do a callout */
    gettimeofday(&now, NULL);
    if ((now.tv_sec - lastcout.tv_sec) >= 60) {
      timertick();
      if ((now.tv_sec - keepalive.tv_sec) >= 120) {
	/* modem is hung or not in StarMode; reinitialize it and kill all
	   handles in handle table*/
	modem_reset(modemfd);
	ask_neighbors();
	keepalive = now;
	keepalive.tv_sec += 30;
      }
      lastcout = now;
    }
    /* process any received packets */
    if (FD_ISSET(tunnelfd, &rfds)) {
      /* got a packet from tunnel */
      plen = read(tunnelfd, txpacket, RICOCHET_MTU);
      ipee = (struct ip *) txpacket;
      /* nnarp for IP destination */
      /* is the destination multicast? */
      if (IN_MULTICAST(ntohl(ipee->ip_dst.s_addr))) {
	/* yes. iterate over neighbors, replicating packet to each */
	nnarp_neighb_reset();
	while ((macdest = nnarp_neighb_iter()) != NULL) {
	  if (ignore.s_addr != INADDR_ANY) {
	    if (macdest->ipdst.s_addr != ignore.s_addr)
	      humr_ip_send(macdest, txpacket, plen);
	  }
	  else
	    humr_ip_send(macdest, txpacket, plen);
	}
      }
      else if ((macdest = nnarp_lookup((lflag ? rtfd_lookup(ipee->ip_dst) :
					ipee->ip_dst))) != NULL)
	/* send IP packet to modem */
	humr_ip_send(macdest, txpacket, plen);
      else
	fprintf(stderr, "lookup of IP packet send to %lx failed\n",
		ipee->ip_dst.s_addr);
    }
    if (FD_ISSET(modemfd, &rfds)) {
      /* got data from serial */
      nbytes = read(modemfd, accptr, MAX_STAR_PLEN);
      buffered += nbytes;
      /* find CRs; they terminate packets */
      for (curr = accptr; curr < (accptr+nbytes); curr++) {
	if (*curr == '\r') {
	  /* this is the end of a star mode packet! */
	  *curr = '\0';
	  /* check the packet's checksum, if it's of sane length */
	  if ((curr-prevptr-4) > 7) {
	    csum = packet_checksum(prevptr, curr-prevptr-4);
	    sscanf(curr-4, "%4x", &rsum);
	    /* compare packet's checksum with that in packet */
	    if (csum == rsum) {
	      /* checksum is good */
	      /* the radio must have been up and in StarMode recently;
		 record the approximate time */
	      keepalive = now;
	      ptype = packet_type(prevptr, curr-prevptr, srcaddr);
	      switch (ptype) {
	      case RICOCHET_PTYPE_RADIO:
		/* unstuff payload */
		UnStuffData(prevptr+strlen(srcaddr)+2, curr, recvpacket,
			    sizeof(recvpacket));
		/* read HUMR header */
		switch(ntohs(recvheader->humr_type)) {
		case HUMR_TYPE_IAM:
		  receive_iam(recvheader->src, *(struct in_addr *)
			      (recvpacket+sizeof(humr_header)), srcaddr);
		  break;
		case HUMR_TYPE_IP:
		  /* it's possible, but unlikely, that we don't have an NNARP
		     entry for the source, if, say, our modem's been reset.
		     make sure we've got the entry we need to reply. */
		  ipee = (struct ip *) (recvpacket + sizeof(humr_header));
#if 0
		  receive_iam(recvheader->src, ipee->ip_src, srcaddr);
#endif
		  /* block until output to tunnel possible */
		  write(tunnelfd, recvpacket + sizeof(humr_header),
			ntohs(ipee->ip_len));
		  break;
		default:
		  fprintf(stderr, "yow! HUMR packet of unknown type %d!\n",
			  ntohs(recvheader->humr_type));
		  break;
		}
		break;
	      case RICOCHET_PTYPE_ATR:
		process_packet_ATR(prevptr);
		break;
	      case RICOCHET_PTYPE_ERR:
		process_packet_ERR(prevptr);
		break;
	      case RICOCHET_PTYPE_INFO:
		/* possibly a new or lost node message */
		process_packet_INFO(prevptr);
		break;
	      case RICOCHET_PTYPE_ACK:
	      case RICOCHET_PTYPE_NULL:
	      default:
		break;
	      }
	    }
	  }
	  buffered -= curr-prevptr+1;
	  prevptr = curr + 1;
	}
      }
      accptr += nbytes;
      /* if next read could take us past end of buffer, move remaining bytes
	 back to start of buffer and adjust pointers accordingly */
      if ((accptr + MAX_STAR_PLEN) > (paccum + sizeof(paccum))) {
	/* bcopying overlapping regions OK */
	bcopy(prevptr, paccum, buffered);
	prevptr = paccum;
	accptr = paccum + buffered;
      }
    }
  }
}
