/*

  Copyright (C) 1999 President and Fellows of Harvard College

  All rights reserved.

  NOTICE: This software is provided "as is", without any warranty,
  including any implied warranty for merchantability or fitness for a
  particular purpose.  Under no circumstances shall Harvard University
  or its faculty, staff, students or agents be liable for any use of,
  misuse of, or inability to use this software, including incidental
  and consequential damages.

  License is hereby given to use, modify, and redistribute this
  software, in whole or in part, for any commercial or non-commercial
  purpose, provided that the user agrees to the terms of this
  copyright notice, including disclaimer of warranty, and provided
  that this copyright notice, including disclaimer of warranty, is
  preserved in the source code and documentation of anything derived
  from this software.  Any redistributor of this software or anything
  derived from this software assumes responsibility for ensuring that
  any parties to whom wuch a redistribution is made are fully aware of
  the terms of this license and disclaimer.

  Author: Brad Karp, Harvard University EECS, September, 1997

*/

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <regex.h>
#include <unistd.h>
#include <termios.h>
#include <sys/ioctl.h>
#include "debug.h"
#include "modem.h"
#include "stuff.h"
#include "humr.h"

/* don't want to pull in all the arp.h stuff */
int add_neighbor(char *);
void hand_flushall(void);

#define escstr sleep(3); write(fd, "+++", 3); sleep(3)
#define recomp_err(a) fprintf(stderr, "can't compile %s\n", a)

/* regex sources */
char src_out_prefix[] = "*&COMMAND*";
char src_neighb_q[] = "AT~LA";
char src_handle[] = "&E[A-F0-9]+";
char src_mac_q[] = "[0-9]+-[0-9]+\r";
char src_info_new[] = "INF_001";
char src_info_lost[] = "INF_002";
char src_err_hremap[] = "ERR_002";

/* compiled regexes for finding tags */
regex_t rx_neighb_q;
regex_t rx_handle;
regex_t rx_mac_q;
regex_t rx_info_new;
regex_t rx_info_lost;
regex_t rx_err_hremap;

/* XXX -- hack: the fd of the modem device */
int modemfd;

extern humr_mac_addr mymac;

void modem_regex_init(void)
{
  if (regcomp(&rx_handle, src_handle, REG_EXTENDED) != 0)
    recomp_err(src_handle);
  if (regcomp(&rx_neighb_q, src_neighb_q, REG_EXTENDED) != 0)
    recomp_err(src_neighb_q);
  if (regcomp(&rx_mac_q, src_mac_q, REG_EXTENDED) != 0)
    recomp_err(src_mac_q);
  if (regcomp(&rx_info_new, src_info_new, REG_EXTENDED) != 0)
    recomp_err(src_info_new);
  if (regcomp(&rx_info_lost, src_info_lost, REG_EXTENDED) != 0)
    recomp_err(src_info_lost);
  if (regcomp(&rx_err_hremap, src_err_hremap, REG_EXTENDED) != 0)
    recomp_err(src_err_hremap);
}

/* reset the modem; put it into a known state where we can issue commands */
void modem_reset(int fd)
{
  struct termios tios;
  char bsetstr[] = "ate0q1s304=115200\r";
  char ctlstr[] = "at&d0&k3\r";
  char stmodestr[] = "atdt**starmode\r";
  char csumstr[] = "ats319=5\r";
  char macqstr[] = "ats305?\r";
  char respbuf[100];
  char *respptr = respbuf;
  int nbytes, foo;
  regmatch_t minfo;

  /* set up proper attributes on line */
  tcgetattr(fd, &tios);
  /* no input or output processing */
  cfmakeraw(&tios);
  /* require one character be read, no time limit */
  tios.c_cc[VMIN] = 1;
  tios.c_cc[VTIME] = 0;
  /* ignore breaks and parity errors on input */
  tios.c_iflag |= (IGNBRK | IGNPAR);
  /* 8 bits, please */
  tios.c_cflag &= ~CSIZE;
  tios.c_cflag |= CS8;
  /* no parity generation or checking */
  tios.c_cflag &= ~PARENB;
  /* ignore carrier */
  tios.c_cflag |= CLOCAL;
  /* set input and output baud rate */
  cfsetspeed(&tios, B115200);
  tcsetattr(fd, TCSANOW | TCSAFLUSH, &tios);
  /* escape; this should exit starmode */
  escstr;
  sleep(1);
  /* try escape sequence again at 57.6 Kbps */
  cfsetspeed(&tios, B57600);
  tcsetattr(fd, TCSANOW | TCSAFLUSH, &tios);
  /* escape; this should exit starmode */
  escstr;
  /* modem will not hear us if it's pegged at 115200; no matter */
  write(fd, bsetstr, sizeof(bsetstr)-1);
  /* give string time to be sent */
  sleep(1);
  /* now jump to 115 Kbps */
  cfsetspeed(&tios, B115200);
  tcsetattr(fd, TCSANOW | TCSAFLUSH, &tios);
  /* send init string again, in case modem was at 115200 all along */
  write(fd, bsetstr, sizeof(bsetstr)-1);
  write(fd, ctlstr, sizeof(ctlstr)-1);
  /* give strings time to be sent */
  sleep(1);
  /* use CTS/RTS handshaking now that we've turned it on; flush input */
  tios.c_cflag |= CRTSCTS;
  tcsetattr(fd, TCSANOW | TCSAFLUSH, &tios);
  /* turn on star mode checksums */
  write(fd, csumstr, sizeof(csumstr)-1);
  /* ask for our local MAC address from modem */
  write(fd, macqstr, sizeof(macqstr)-1);
  bzero(respbuf, sizeof(respbuf));
  /* read response */
  while (regexec(&rx_mac_q, respbuf, 1, &minfo, 0) &&
	 ((respptr-respbuf) < 80)) {
    nbytes = read(fd, respptr, 20);
    respptr += nbytes;
  }
  if ((minfo.rm_eo - minfo.rm_so) == 10) {
    /* found it! */
    sscanf(respbuf+minfo.rm_so, "%4d", &foo);
    mymac.hi = foo;
    sscanf(respbuf+minfo.rm_so+5, "%4d", &foo);
    mymac.lo = foo;
    if (dmask & DMASK_MINIT) {
      fprintf(stderr, "my MAC address is %d-%d\n", mymac.hi, mymac.lo);
      fflush(stderr);
    }
  }
  else {
    if (dmask & DMASK_MINIT) {
      respbuf[sizeof(respbuf)-1] = '\0';
      fprintf(stderr, "got %s from modem after MAC addr query\n",
	      respbuf);
      fflush(stderr);
    }
    fprintf(stderr, "can't find my MAC address! exiting...\n");
    exit(1);
  }
  /* flush input */
  tcsetattr(fd, TCSANOW | TCSAFLUSH, &tios);
  /* put modem into star mode */
  write(fd, stmodestr, sizeof(stmodestr)-1);
  /* flush all handles */
  hand_flushall();
}

/* ask for the neighbors list */
void ask_neighbors(void)
{
  unsigned char combuf[80];
  int thesum, len;

  bzero(combuf, sizeof(combuf));
  strcpy(combuf, src_out_prefix);
  strcat(combuf, src_neighb_q);
  len = strlen(src_out_prefix)+strlen(src_neighb_q);
  thesum = packet_checksum(combuf, len);
  sprintf(combuf+len, "%04X", thesum);
  write(modemfd, combuf, len+4);
  write(modemfd, "\r", 1);
  if (dmask & DMASK_NEIGH) {
    fprintf(stderr, "requested neighbor list\n");
    fflush(stderr);
  }
}

/* given a starmode packet of length len, classify it based on source.
   stuff the source handle or modem ID into psrc if it's from a modem */
int packet_type(u_char *buf, int len, u_char *psrc)
{
  char tmp[20];
  char *cptr;

  if (len == 0)
    return(RICOCHET_PTYPE_NULL);
  if (*buf++ != '*')
    return(RICOCHET_PTYPE_NULL);
  len--;
  len = min(len, sizeof(tmp)-1);
  /* grab source field */
  cptr = tmp;
  while ((*buf != '*') && len--) {
    *cptr++ = *buf++;
  }
  *cptr = '\0';
  if (strcmp(tmp, "!MODEM") == 0)
    return(RICOCHET_PTYPE_ATR);
  else if (strcmp(tmp, "!ACK") == 0)
    return(RICOCHET_PTYPE_ACK);
  else if (strcmp(tmp, "!INFO") == 0)
    return(RICOCHET_PTYPE_INFO);
  else if (strcmp(tmp, "!ERROR") == 0)
    return(RICOCHET_PTYPE_ERR);
  else {
    strcpy(psrc, tmp);
    return(RICOCHET_PTYPE_RADIO);
  }
}

/* process a null-terminated starmode response to a modem command;
   return 0 on success, -1 on failure */
int process_packet_ATR(u_char *buf)
{
  regmatch_t minfo;
  char curr_handle[50];

  /* was this a request for neighbors? */
  if (regexec(&rx_neighb_q, buf, 1, &minfo, 0) == 0) {
    buf += minfo.rm_eo;
    /* found neighbor request; grab handles */
    while (regexec(&rx_handle, buf, 1, &minfo, 0) == 0) {
      /* got another handle */
      strncpy(curr_handle, buf+minfo.rm_so, minfo.rm_eo-minfo.rm_so);
      curr_handle[minfo.rm_eo-minfo.rm_so] = '\0';
      if (dmask & DMASK_NEIGH) {
	fprintf(stderr, "neighbor found: %s\n", curr_handle);
	fflush(stderr);
      }
      add_neighbor(curr_handle);
      buf += minfo.rm_eo;
    }
    return(0);
  }
  return(-1);
}

int process_packet_INFO(u_char *buf)
{
  regmatch_t minfo;
  char curr_handle[50];

  /* is this a new neighbor packet? */
  if (regexec(&rx_info_new, buf, 1, &minfo, 0) == 0) {
    buf += minfo.rm_eo;
    /* found new neighbor */
    if (regexec(&rx_handle, buf, 1, &minfo, 0) == 0) {
      /* grabbed a handle */
      strncpy(curr_handle, buf+minfo.rm_so, minfo.rm_eo-minfo.rm_so);
      curr_handle[minfo.rm_eo-minfo.rm_so] = '\0';
      if (dmask & DMASK_NEIGH) {
	fprintf(stderr, "neighbor found: %s\n", curr_handle);
	fflush(stderr);
      }
      add_neighbor(curr_handle);
      return(0);
    }
  }
  else if (regexec(&rx_info_lost, buf, 1, &minfo, 0) == 0) {
    buf += minfo.rm_eo;
    /* lost a neighbor */
    /* XXX - for now, do nothing...possibly never do anything */
    if (dmask & DMASK_NEIGH) {
      fprintf(stderr, "lost neighbor (no action): %s\n", buf);
      fflush(stderr);
    }
    return(0);
  }
  return(-1);
}

int process_packet_ERR(u_char *buf)
{
  regmatch_t minfo;
  char curr_handle[50];
  int hand_delete(char *);

  if (regexec(&rx_err_hremap, buf, 1, &minfo, 0) == 0) {
    buf += minfo.rm_eo;
    /* we sent to a MAC addr we previously knew only by a handle in this
       StarMode session */
    if (regexec(&rx_handle, buf, 1, &minfo, 0) == 0) {
      /* try to delete the handle, in case we have it in the handle table */
      strncpy(curr_handle, buf+minfo.rm_so, minfo.rm_eo-minfo.rm_so);
      curr_handle[minfo.rm_eo-minfo.rm_so] = '\0';
      if (dmask & DMASK_NEIGH) {
	fprintf(stderr, "flush handle: %s\n", curr_handle);
	fflush(stderr);
      }
      hand_delete(curr_handle);
      return(0);
    }
  }
  return(-1);
}

/* compute starmode checksum over complete packet, assumed to have no trailing
   checksum */
int packet_checksum(u_char *buf, int len)
{
  unsigned short checksum = 0;

  if (len <= 7)
    return (CHECKSUM_INVALID);
  while (len--)
    checksum += *buf++;
  return (checksum);
}

int ishandle(char *addr)
{
  if ((strlen(addr) > 2) && (addr[0] == '&') && (addr[1] == 'E'))
    return 1;
  else
    return 0;
}
